Require Import CertifiedBSML.Primitives.Specification.
Require Import CertifiedBSML.Primitives.Properties.

Require Import LIFO.Vector.Main.
Require Import CertifiedBSML.Primitives.RealisationVectors.

Require CertifiedBSML.Support.ParSig.
Require CertifiedBSML.StandardLibrary.Base.
Require CertifiedBSML.StandardLibrary.Comm.
Require Arith.

Require Import Strings.String.

(** Module Type Enabling to parametrize modules with the number of processors *)
Module Type BSML_PARAMETER.
  Parameter numberOfProcessors : { n : nat | 0 < n }.
End BSML_PARAMETER.

(* Module Type Enabling to parametrize modules by a printing function*)
Module Type TEST_PARAMETERS.
(** should print a boolean result, giving a description, 
   the number of procs sequentially simulated
   and the boolean value
   *)
  Parameter print_result : string ->nat -> bool -> unit.
End TEST_PARAMETERS.




(** Module containing functions to be tested  *)
Module TestingFunction (param: BSML_PARAMETER)
  (Bsml : CertifiedBSML.Primitives.Specification.PRIMITIVES with Definition bsp_p:= proj1_sig param.numberOfProcessors)
  (Props : CertifiedBSML.Primitives.Properties.TYPE Bsml)
  (ParSig : CertifiedBSML.Support.ParSig.TYPE Bsml Props)
  (Base : CertifiedBSML.StandardLibrary.Base.TYPE Bsml Props ParSig)
  (Comm : CertifiedBSML.StandardLibrary.Comm.TYPE Bsml Props ParSig Base).

  Import Bsml Props ParSig Base Comm.
  
  Program Definition par_to_seq A (V: par A) :=  
    map (proj V) (@inSig _ _ _  (seq param.numberOfProcessors 0) _).
  Next Obligation.
    intros A V a H.
    apply inSeq0Lt  in H.
    unfold bsp_p.
    assumption.
  Qed.
  
  Program Definition next_this := apply (mkpar (fun _ i => S (proj1_sig i))) this.
  Program Definition shifted_right_next_this:= shift_right next_this.
  Program Definition shifted_left_next_this:= shift_left next_this.

Implicit Arguments par_to_seq [A].  
End TestingFunction.
  

(** Module containing functions to be tested implemented using sequential realisation
   And a proof that in this implementation par_to_seq is the identity
   *)
Module SequentialTest (param : BSML_PARAMETER).

  Module BsmlImplementation:=BsmlImplementation  param.
  Module Properties:=CertifiedBSML.Primitives.Properties.Make BsmlImplementation.
  Module ParSig:=CertifiedBSML.Support.ParSig.Make BsmlImplementation Properties.
  Module Base:=  CertifiedBSML.StandardLibrary.Base.Make BsmlImplementation  Properties ParSig.
  Module Comm:=  CertifiedBSML.StandardLibrary.Comm.Make BsmlImplementation  Properties ParSig Base.
  Include TestingFunction param BsmlImplementation Properties ParSig Base Comm.
  (* TODO   we should prove that par_to_seq is identity in sequential implementation *)
  Lemma par_to_seq_idem A (V : BsmlImplementation.par A) : par_to_seq  V = V.
  Proof.
    unfold par_to_seq.
    revert V.
  Admitted.
End SequentialTest.




(** list of sequential values computed to be tested against parallel results 
   this list should match the list of testing functions !
   *)
Module Type ComputedSequentialTest (param : BSML_PARAMETER).
  Module seqTest:= SequentialTest param.
  Import seqTest. 

  Parameter next_this_computed : BsmlImplementation.par nat.
  Parameter shifted_right_next_this_computed : BsmlImplementation.par nat.
  Parameter shifted_left_next_this_computed : BsmlImplementation.par nat.

End ComputedSequentialTest.




(** equality has to deal with different size of vectors as number  of procs chosen when computing sequentially  is not alway the nubmer of procs parallely available  *)
Program Fixpoint vector_beq A B  n  m a_b_beq (v1 :vector A n) (v2 :vector B m) := 
  match v1 with 
    Vnil => true
    | Vcons a n v => match v2 with
                     Vnil => false
                     | Vcons a' n' v' => andb (a_b_beq a a') (vector_beq A B n n'  a_b_beq v v')
                   end
  end.
Implicit Arguments vector_beq [A B n m ].

(** This module take a module containing  sequentially computed values and test them against parallely computed values  *)
Module  Tester
  (param : BSML_PARAMETER)
  (ComputedSequentialTest :  ComputedSequentialTest param)
  (BsmlNat : CertifiedBSML.Primitives.Specification.PRIMITIVES )
  (Printer : TEST_PARAMETERS).
  
  Module Props := CertifiedBSML.Primitives.Properties.Make BsmlNat.
  Module ParSig := CertifiedBSML.Support.ParSig.Make BsmlNat Props.
  Module Base := CertifiedBSML.StandardLibrary.Base.Make BsmlNat Props ParSig.
  Module Comm := CertifiedBSML.StandardLibrary.Comm.Make BsmlNat Props ParSig Base.

  Import Printer.
   
  Module TrueParams.
    Program Definition numberOfProcessors : { n : nat | 0 < n }:=BsmlNat.bsp_p.
    Solve Obligations using (intros; auto with arith bsml).
  End TrueParams.
  
  (** parallel functions  *)
  Module ParallelTestingFunction :=TestingFunction TrueParams BsmlNat Props ParSig Base Comm.

  (** testing each functions  *)
  
  Definition SeqParProcOk :=  Peano_dec.eq_nat_dec  BsmlNat.bsp_p (proj1_sig param.numberOfProcessors).
  
  Program Definition testing_next_this := 
    if SeqParProcOk then
      print_result  "vector of processor :"  param.numberOfProcessors
      (vector_beq (EqNat.beq_nat)
        ComputedSequentialTest.next_this_computed
        (ParallelTestingFunction.par_to_seq  ParallelTestingFunction.next_this)
      )
      else
      tt.
  
  Program Definition testing_shifted_right_next_this := 
    if SeqParProcOk then
      print_result  "circular rigth  shifting of processor :" param.numberOfProcessors
      (vector_beq (EqNat.beq_nat)
        ComputedSequentialTest.shifted_right_next_this_computed
        (ParallelTestingFunction.par_to_seq  ParallelTestingFunction.shifted_right_next_this)
      )
    else
      tt.
  
  Program Definition testing_shifted_left_next_this := 
    if SeqParProcOk then
      print_result  "circular rigth  shifting of processor :" param.numberOfProcessors
      (vector_beq (EqNat.beq_nat)
        ComputedSequentialTest.shifted_left_next_this_computed
        (ParallelTestingFunction.par_to_seq  ParallelTestingFunction.shifted_left_next_this)
      )
      else
        tt.

End Tester.


(** now instantiating modules for different numbers of processors  *)
Module Bsml_1.
  Program Definition numberOfProcessors : { n : nat | 0 < n }:=1.
End Bsml_1.


Module  ComputedSequentialTest_1 : (ComputedSequentialTest Bsml_1).
  Module seqTest:= SequentialTest Bsml_1.
  Import seqTest. 

   Definition  next_this_computed : BsmlImplementation.par nat := 
    Eval compute in proj1_sig next_this.
  
   Definition shifted_right_next_this_computed : BsmlImplementation.par nat  := 
    Eval compute in  proj1_sig (shifted_right_next_this).

  Definition shifted_left_next_this_computed : BsmlImplementation.par nat := 
    Eval compute in proj1_sig(shifted_left_next_this).
End ComputedSequentialTest_1.


Module Bsml_2.
  Program Definition numberOfProcessors : { n : nat | 0 < n }:=2.
  Next Obligation.
    auto with arith.
  Qed.
End Bsml_2.

Module  ComputedSequentialTest_2 : ComputedSequentialTest Bsml_2.
  Module seqTest:= SequentialTest Bsml_2.
  Import seqTest. 

  Definition  next_this_computed : BsmlImplementation.par nat := 
    Eval compute in proj1_sig(next_this).
  
Definition shifted_right_next_this_computed : BsmlImplementation.par nat  := 
    Eval compute in proj1_sig(shifted_right_next_this).

  Definition shifted_left_next_this_computed : BsmlImplementation.par nat := 
    Eval compute in proj1_sig(shifted_left_next_this).
End ComputedSequentialTest_2.


Module Bsml_4.
  Program Definition numberOfProcessors : { n : nat | 0 < n }:=4.
  Next Obligation.
    auto with arith.
  Qed.
End Bsml_4.

Module  ComputedSequentialTest_4 : ComputedSequentialTest Bsml_4.
  Module seqTest:= SequentialTest Bsml_4.
  Import seqTest. 

  Definition  next_this_computed : BsmlImplementation.par nat := 
    Eval compute in proj1_sig(next_this).
  
Definition shifted_right_next_this_computed : BsmlImplementation.par nat  := 
    Eval compute in proj1_sig(shifted_right_next_this).

  Definition shifted_left_next_this_computed : BsmlImplementation.par nat := 
    Eval compute in proj1_sig(shifted_left_next_this).
End ComputedSequentialTest_4.


Module Bsml_8.
  Program Definition numberOfProcessors : { n : nat | 0 < n }:=8.
  Next Obligation.
    auto with arith.
  Qed.
End Bsml_8.

Module ComputedSequentialTest_8 : ComputedSequentialTest Bsml_8.
  Module seqTest:= SequentialTest Bsml_8.
  Import seqTest. 

  Definition  next_this_computed : BsmlImplementation.par nat := 
    Eval compute in proj1_sig(next_this).
  
Definition shifted_right_next_this_computed : BsmlImplementation.par nat  := 
    Eval compute in proj1_sig(shifted_right_next_this).

  Definition shifted_left_next_this_computed : BsmlImplementation.par nat := 
    Eval compute in proj1_sig(shifted_left_next_this).
End ComputedSequentialTest_8.


Module Bsml_7.
  Program Definition numberOfProcessors : { n : nat | 0 < n }:=7.
  Next Obligation.
    auto with arith.
  Qed.
End Bsml_7.

Module ComputedSequentialTest_7 : ComputedSequentialTest Bsml_7.
  Module seqTest:= SequentialTest Bsml_7.
  Import seqTest. 

  Definition  next_this_computed : BsmlImplementation.par nat := 
    Eval compute in proj1_sig(next_this).
  
Definition shifted_right_next_this_computed : BsmlImplementation.par nat  := 
    Eval compute in proj1_sig(shifted_right_next_this).

  Definition shifted_left_next_this_computed : BsmlImplementation.par nat := 
    Eval compute in proj1_sig(shifted_left_next_this).
End ComputedSequentialTest_7.




Module Tester_1 := Tester Bsml_1 ComputedSequentialTest_1.
Module Tester_2 := Tester Bsml_2 ComputedSequentialTest_2.
Module Tester_4 := Tester Bsml_4 ComputedSequentialTest_4.
Module Tester_8 := Tester Bsml_8 ComputedSequentialTest_8.
Module Tester_7 := Tester Bsml_7 ComputedSequentialTest_7.
