open Bsml
open Stdlib
open Base
open Comm
open Unix
open DataGenerator

open Datatypes
open List0

  (** version of program to be used  *)
  type versions = Coq|Direct|CoqWarming|DirectWarming|Defunct|DefunctWarming|All|Both

  let versions_of_string s = 
    let lows = String.lowercase s in
    match  lows with
	"coq" -> Coq
      | "direct" -> Direct
      | "defunct" -> Defunct
      | "coqwarming" -> CoqWarming
      | "directwarming" -> DirectWarming
      | "defunctwarming" -> DefunctWarming
      | "both" -> Both
      | "all" -> All
      |  _ -> failwith "conversion error, not an element of type version (benchmarking.ml)"

  let string_of_versions v = 
    match  v with
	Coq   ->   "Coq"    
      | Direct ->    "Direct" 
      | Defunct -> "Defunct"
      | CoqWarming   ->   "CoqWarming"    
      | DirectWarming ->    "DirectWarming" 
      | DefunctWarming -> "DefunctWarming"
      | Both   ->    "Both"   
      | All -> "All"
	  

let human_readable = ref false

let argshift = ref 0

let incremental_mode = ref false

let incremental_init = ref 100

let incremental_increment = ref 10

let program_name = ref "unkown"

let current_version = ref Both

let reuse_data = ref false

let set_incremental_mode  init_size increment = 
  incremental_mode:= true;
  incremental_init := init_size;
  incremental_increment:=increment


let set_human_readable () =
  human_readable := true
let set_reusedata () =
  reuse_data:=true
let doat0 f = mkpar (fun i -> if i = 0 then f () else  ())

let compactGCsilently ()=
  begin
    print_string "SILENT GC ...";
    let verbosity = ((Gc.get()).Gc.verbose) in
      (Gc.set { (Gc.get()) with Gc.verbose = 0x000 };
       Gc.compact();
       Gc.set { (Gc.get()) with Gc.verbose =  verbosity};
      );
      print_string "END ";
  end

    
let print_separator  s =
  if (!human_readable) then
    print_string s
  else
    print_string " ; "
let gctiming = ref 0.

(** print computation time log for program [program_name] applied to data of size [size]    *)
let print_time program_name  size  time =
  begin
    let cents = int_of_float (time*.100.) in
    let sec = float_of_int (cents mod 6000)/.100. in
    let seconds = cents / 100 in
    let minutes = seconds /60 in
    let min = minutes mod 60 in
    let hours = minutes /60 in
    let h = hours mod 60 in
    let days = hours /24 in
      print_string program_name; print_separator  " computing over ";
      print_int bsp_p ; print_separator  " processor and ";
      print_int size; print_separator " data in time ";
      print_float time ; print_separator " human readable time ";
      if days > 0 then ( print_int days ; print_string "days " );
      if h>0 then ( print_int h  ; print_string "h ");
      if min>0 then ( print_int min ; print_string "min " );
      if sec>0. then (print_float sec ; print_string "sec ");
      if time < 0.01 then  (print_float time ; print_string "sec ");
      print_separator "gc time"; print_float !gctiming ;
      print_string "\n";
      flush Pervasives.stdout
  end

(** print computation time log for program [program_name] applied to data of size [size]    *)
let print_time_normalize program_name  size  time   allocated_size gc_num gc_time =
  begin
    print_string (program_name ^(string_of_versions !current_version));
    print_separator  " executing on ";
    if !incremental_mode || !human_readable then
      begin
	print_int bsp_p ;
	print_separator  " processors with data of size ";	
	print_int size; 
	print_separator  " in time ";	
      end;    
    let cents = int_of_float (time*.100.) in
    let sec = float_of_int (cents mod 6000)/.100. in
    let seconds = cents / 100 in
    let minutes = seconds /60 in
    let min = minutes mod 60 in
    let hours = minutes /60 in
    let h = hours mod 60 in
    let days = hours /24 in
    print_float time ; 
    if (!human_readable) then
      begin
	print_separator " human readable time ";
	if days > 0 then ( print_int days ; print_string "days " );
	if h>0 then ( print_int h  ; print_string "h ");
	if min>0 then ( print_int min ; print_string "min " );
	if sec>0. then (print_float sec ; print_string "sec ");
	if time < 0.01 then  (print_float time ; print_string "sec ");
      end;
    print_separator " allocated words "; print_float allocated_size;
    print_separator " number of garbage collection "; print_int gc_num ;
    print_separator " gc time"; print_float !gctiming ;
    if (!human_readable || !incremental_mode) then
      print_string "\n";
    flush Pervasives.stdout
  end


let print_timing  program_name  size   =       
  ignore(bsml_print (print_time program_name  size) 0 (fold_direct max (get_cost())))

let print_step log =
  begin
    ignore (bsml_print  (output_string Pervasives.stderr) 0 (replicate log));
    flush Pervasives.stderr
  end

let rec string_of_list string_of_element separator = function
    | []  -> ""
    | h::[] -> string_of_element h
    | h:: t      -> (string_of_element h)^separator^
       (string_of_list string_of_element separator t)

let rec string_of_Coq_list string_of_element separator = string_of_list string_of_element separator
(* version for old extraction of Coq list  *)
  (* function *)
  (*   | Coq_nil  -> "" *)
  (*   | Coq_cons (h,Coq_nil) -> string_of_element h *)
  (*   | Coq_cons (h,t)      -> (string_of_element h)^separator^ *)
  (*      (string_of_Coq_list string_of_element separator t) *)


let string_of_data =  string_of_float

let print_data_coq_lst = parprint (fun l ->(print_string (string_of_Coq_list (fun c ->string_of_data c)  " ; " l)))

let print_data_lst = parprint (fun l ->(print_string (string_of_list (fun c ->string_of_data c)  " ; " l)))

let data_size_warning size =
  let local_size =size/bsp_p in
  if (local_size > 200000) then print_step ("/!\ warning /!\  local list size ("^ (string_of_int local_size) ^" exceed stack limit, risk of overflow ")


(** perform [iteration] benchmark of the function [f] applied to [data] of size [size], naming it  [program_name] in the logging output   *)
let rec benchmark  f data program_name  size iteration   =
  begin
    if iteration >0 then (
      start_timing();
      Gc.compact();
      stop_timing();
      ignore(bsml_print (fun x -> gctiming:=x) 0 (fold_direct max (get_cost())));
      start_timing();
      f size data;
      stop_timing();
      ignore(bsml_print (print_time program_name  size) 0 (fold_direct max (get_cost())));
      benchmark f data program_name  size (iteration-1);
    )
  end

(** perform [iteration] benchmark of the function [f] for each size given  in  size_list.
    data are randomly generated using balanced_random_data_par
*)
let rec benchmark_list local_data_generator f max_data program_name  size_list  iteration =
  match size_list with
    | [] -> ();
    | size::l ->(
        let data    = balanced_random_data_par local_data_generator max_data  (size/bsp_p) in
          benchmark f data program_name  size iteration ;
          benchmark_list  local_data_generator f max_data program_name  l iteration
      )


(** default benchmark perform 20 call to [f] for sizes 1000 10000 100000 and 1000000.
*)
let one_size_default_benchmark local_data_generator f max_data program_name size : unit =
  print_step ("starting  benchmark for "^program_name^"...\n");
  benchmark_list local_data_generator f max_data program_name  [size] 4;
  print_step ("end of benchmark for "^program_name^".\n")


let unref v = << ! $v$ >>

  let one_benchmark_reuse iter  program data size =
  (* performing major collection *)
    Gc.full_major();
    let gc_cost= << ref 0. >> in 
    let cost = << ref 0. >> in
    let start_minor_collection_vector= << let stat =Gc.stat() in stat.Gc.minor_collections>> in
    begin
      Bsml.start_timing();
      let data' = ref data in
      for i=1 to iter do
	data':=program  size !data';
	Bsml.stop_timing();
	let cost_vec=Bsml.get_cost() in
	ignore(<< $cost$:=( $cost_vec$ +. ! ($cost$)) >>);
	Bsml.start_timing();
	Gc.full_major();
	Bsml.stop_timing();
	ignore( << $gc_cost$ := $Bsml.get_cost()$ +.  ! $gc_cost$>>);
      Bsml.start_timing();
      done; 
      Bsml.stop_timing();
      ignore(<< $cost$:= $fold_direct max  (unref cost)$ >>);
      let word_vector= fold_direct max  << let stat =Gc.stat() in stat.Gc.minor_words +.  stat.Gc.major_words -. stat.Gc.promoted_words>> in
      let minor_collection_vector= fold_direct max  << let stat =Gc.stat() in stat.Gc.minor_collections>> in
      Bsml.start_timing();
      Gc.full_major();
      Bsml.stop_timing();
      ignore( << $gc_cost$ := $Bsml.get_cost()$ +.  ! $gc_cost$>>);
      (*  *)
      ignore (<< if ( $this$ = 0 ) then 
	  print_time_normalize !program_name size   ! $cost$ $word_vector$   ($minor_collection_vector$ - $start_minor_collection_vector$) $fold_direct max   (unref gc_cost)$   
	else ()
	  >> )
    end

  let one_benchmark_no_reuse iter  program data size =
  (* performing major collection *)
    Gc.full_major();
    let gc_cost= << ref 0. >> in 
    let cost = << ref 0. >> in
    let start_minor_collection_vector= << let stat =Gc.stat() in stat.Gc.minor_collections>> in
    begin
      Bsml.start_timing();
      for i=1 to iter do
	program size data;
	Bsml.stop_timing();
	let cost_vec=Bsml.get_cost() in
	ignore(<< $cost$:=( $cost_vec$ +. ! ($cost$)) >>);
	Bsml.start_timing();
	Gc.full_major();
	Bsml.stop_timing();
	ignore( << $gc_cost$ := $Bsml.get_cost()$ +.  ! $gc_cost$>>);
      Bsml.start_timing();
      done; 
      Bsml.stop_timing();
      ignore(<< $cost$:= $fold_direct max  (unref cost)$ >>);
      let word_vector= fold_direct max  << let stat =Gc.stat() in stat.Gc.minor_words +.  stat.Gc.major_words -. stat.Gc.promoted_words>> in
      let minor_collection_vector= fold_direct max  << let stat =Gc.stat() in stat.Gc.minor_collections>> in
      Bsml.start_timing();
      Gc.full_major();
      Bsml.stop_timing();
      ignore( << $gc_cost$ := $Bsml.get_cost()$ +.  ! $gc_cost$>>);
      (*  *)
      ignore (<< if ( $this$ = 0 ) then 
	  print_time_normalize !program_name size   ! $cost$ $word_vector$   ($minor_collection_vector$ - $start_minor_collection_vector$) $fold_direct max   (unref gc_cost)$   
	else ()
	  >> )
    end

  let one_benchmark iter  program data size =
    if !reuse_data then 
      one_benchmark_reuse iter  program data size  
    else 
      one_benchmark_no_reuse iter  program data size 


let benchmarking   program_coq coq_data_generator   program_direct   data_generator  program_defunct =  
  let  _  = try match String.lowercase (Bsml.argv.(1)) with "h"|"human" -> set_human_readable (); incr argshift ; with _ ->() in
  let  _  = try match String.lowercase (Bsml.argv.(1+ !argshift)) with "r"|"reuse" -> set_reusedata (); incr argshift ; with _ ->() in
  let version = try let res= versions_of_string (Bsml.argv.(1+ !argshift))in incr argshift;res with _ -> Both in
  let iterNumber = try let n= int_of_string (Bsml.argv.(1+ !argshift)) in incr argshift; n with _ -> 0 in
  let iter = try int_of_string (Bsml.argv.(1+ !argshift)) with _ -> 200 in
  let size = try int_of_string (Bsml.argv.(2+ !argshift)) with _ -> 10000 in
  let ram_size = try int_of_string (Bsml.argv.(3+ !argshift)) with _ -> 10 in
  let _ = try let init_size = int_of_string (Bsml.argv.(4+ !argshift))  in
	      let increment = try int_of_string (Bsml.argv.(5+ !argshift)) with _ -> 10 in
	      set_incremental_mode init_size increment
    with _ -> ()      
  in
  let i = ref size;
  in
  begin  
    print_step  ("# version :"^(string_of_versions version)^", iter : "^(string_of_int iter)^(if !human_readable then ", human readable" else "")^
		    (if !reuse_data then ", reusing data" else "")^
		    ", size :"^( string_of_int size)^
		    ", ram size :"^(string_of_int ram_size)^
		    ", arg shift :"^(string_of_int !argshift)^"\n") ;
    
    Gc.set { (Gc.get ()) with Gc.minor_heap_size = 1024*1024*1024*ram_size };
    Gc.set { (Gc.get ()) with Gc.space_overhead = 99};
    if !incremental_mode then
      begin
	i:=!incremental_init ;  
	(* warming memory *)
	
	(match version with 
	    Coq     ->   current_version:= CoqWarming;  one_benchmark iter program_coq  (DataGenerator.balanced_par_coq coq_data_generator !i) !i
	   |  Direct ->   current_version:=DirectWarming; one_benchmark iter program_direct (DataGenerator.balanced_par data_generator !i) !i
	   |  Defunct ->   current_version:=DefunctWarming; one_benchmark iter program_defunct (DataGenerator.balanced_par_coq coq_data_generator !i) !i
	   | Both ->
	     if (iterNumber mod 2 = 1) then 
	       begin
		 current_version:= CoqWarming;  one_benchmark iter program_coq  (DataGenerator.balanced_par_coq coq_data_generator !i) !i;
		 current_version:=DirectWarming; one_benchmark iter program_direct (DataGenerator.balanced_par data_generator !i) !i;
	       end
	     else
	       begin
		 current_version:=DirectWarming; one_benchmark iter program_direct (DataGenerator.balanced_par data_generator !i) !i;
		 current_version:= CoqWarming;  one_benchmark iter program_coq  (DataGenerator.balanced_par_coq coq_data_generator !i) !i;
	       end
	   | All ->
	     if (iterNumber mod 3 = 1) then 
	       begin
		 current_version:= CoqWarming;  one_benchmark iter program_coq  (DataGenerator.balanced_par_coq coq_data_generator !i) !i;
		 current_version:=DirectWarming; one_benchmark iter program_direct (DataGenerator.balanced_par data_generator !i) !i;
		 current_version:=DefunctWarming; one_benchmark iter program_defunct  (DataGenerator.balanced_par_coq coq_data_generator !i) !i;
	       end
	     else
	     if (iterNumber mod 3 = 2) then 	       
	       begin
		 current_version:=DirectWarming; one_benchmark iter program_direct (DataGenerator.balanced_par data_generator !i) !i;
		 current_version:=DefunctWarming; one_benchmark iter program_defunct  (DataGenerator.balanced_par_coq coq_data_generator !i) !i;
		 current_version:= CoqWarming;  one_benchmark iter program_coq  (DataGenerator.balanced_par_coq coq_data_generator !i) !i;
	       end
	     else
	       begin
		 current_version:=DefunctWarming; one_benchmark iter program_defunct  (DataGenerator.balanced_par_coq coq_data_generator !i) !i;
		 current_version:= CoqWarming;  one_benchmark iter program_coq  (DataGenerator.balanced_par_coq coq_data_generator !i) !i;
		 current_version:=DirectWarming; one_benchmark iter program_direct (DataGenerator.balanced_par data_generator !i) !i;
	       end
	)
      end
    ;
      
    while !i <= size do (* i is initialized to size if not in incremental mode *)
      (match version with 
	  Coq | Both|All    ->   current_version:= Coq;  one_benchmark iter program_coq  (DataGenerator.balanced_par_coq coq_data_generator !i) !i
	| _  -> ()
      )
      ;
      (match version with 
	  Direct| Both|All ->   current_version:=Direct; one_benchmark iter program_direct (DataGenerator.balanced_par data_generator !i) !i
	  | _ -> ())
      ;
      (match version with 
	  Defunct|All -> current_version:=Defunct; one_benchmark iter program_defunct  (DataGenerator.balanced_par_coq coq_data_generator !i) !i;
	  | _ -> ())
      ;
      i:= !i * !incremental_increment
    done
  end
