Require Import Omega. 
Require Import Arith.
Require Import NPeano.
Require Import List.
Require Import LIFO.List.InSig.
Require Import LIFO.List.Misc.
Require Import LIFO.List.SeqNat.
Require Import LIFO.List.Length.
Require Import LIFO.Nat.BoundedNat.
Require Import CertifiedBSML.BsmlSpec.

Set Implicit Arguments.

Module Type TYPE  
  (BsmlPrimitives: PRIMITIVES).

  Import BsmlPrimitives.

  Definition lmkpar (A:Type)(f:processor->A) := 
    proj2_sig (mkpar f).

  Definition lapply(A B:Type)(vf:par (A->B))(vv:par A):=
    proj2_sig (apply vf vv).

  Definition lproj(A:Type)(v:par A):=
    proj2_sig (proj v).

  Definition lput(A:Type)(v:par(processor->A)):=
    proj2_sig (put v).

  Hint Rewrite lmkpar lapply lproj lput : bsml.

  Lemma bsp_pSMaxPid: 
    exists maxPid:nat, bsp_p = S maxPid.
  Proof.
    case_eq(bsp_p). 
       intro Hbsp_p; set(Hp:= bsp_pNotZero); rewrite Hbsp_p in *; contradict Hp; omega.
       intro maxPid; exists maxPid; trivial.
  Qed.

  Program Definition firstProcessor : processor := 0.
  Next Obligation.
    destruct(bsp_pSMaxPid) as [maxPid HmaxPid];
      rewrite HmaxPid; 
        intuition.
  Qed.

  Program Definition lastProcessor : processor := bsp_p - 1.
  Next Obligation.
    destruct(bsp_pSMaxPid) as [maxPid HmaxPid]; rewrite HmaxPid; omega.
  Qed.

  Program Definition 
    processorOfNatMod (n:nat): processor := 
    n mod bsp_p.
  Next Obligation.
    apply mod_upper_bound;
      destruct(bsp_pSMaxPid) as [maxPid HmaxPid]; rewrite HmaxPid; omega.
  Defined.

  Hint Unfold listProj: bsml_properties.
  Hint Rewrite inSigProj : bsml_properties.

  Lemma processorOfNatModLtBspPId : 
    forall(n:nat)(H:n < bsp_p),
      processorOfNatMod n = exist (fun n=>n<bsp_p) n H.
  Proof.
    intros n H;
      unfold processorOfNatMod;
        apply proj1_sigBoundedNatInjective; simpl;
          apply NZDivP.mod_small; omega.
  Qed.

  Program Definition processors : { l : list processor | listProj l = seq 0 bsp_p} := 
    inSig (fun i=>i<bsp_p) (seq 0 bsp_p) (inSeq0Lt bsp_p). 
  Next Obligation.
    autorewrite with bsml_properties; auto.
  Qed.

  Program Definition processorsLength: 
    List.length processors = bsp_p.
  Proof.
    unfold processors; autorewrite with length; unfold processor; 
      autorewrite with length; trivial.
  Qed.

  Hint Rewrite processorsLength : length.

  Program Definition inProcessors : 
    forall i:processor, In i processors.
  Proof.
    intro i; unfold processors; simpl; destruct i; 
      apply inInSig1; apply inSeq; split; omega.
  Qed.

  Lemma processorsBeginsWithFirstProcessor:
      proj1_sig processors = firstProcessor :: (tail (proj1_sig processors)).
  Proof.
    destruct bsp_pSMaxPid as [maxPid HmaxPid].
    assert(seq 0 bsp_p = 0::seq 1 maxPid) as Hseq by (rewrite HmaxPid; trivial).
    pose(Hlen := processorsLength); generalize  Hlen; clear Hlen; intro Hlen.
    case_eq (proj1_sig processors). 
      intro Hdef; contradict Hlen; rewrite Hdef, HmaxPid; simpl; omega.
      intros first tail Hdef; 
        destruct processors as [processorList Hprocessors]; simpl in Hdef;
          clear Hlen; rewrite Hdef in Hprocessors; simpl in Hprocessors; 
            rewrite Hseq in Hprocessors; inversion Hprocessors;
              f_equal; apply proj1_sigBoundedNatInjective;
                assumption.
  Qed.

  Program Definition processorsUpTo (i:processor) : list processor := 
    inSig (fun n:nat=>n<bsp_p) (seq 0 (S i)) _.
  Next Obligation.
    destruct H.
      subst; apply bsp_pNotZero.
      destruct i as [i Hi]; simpl in *; apply inSeqLt in H; omega.
  Defined.

  Lemma processorsUpToApp:
    forall (n:nat)(H:S n<bsp_p), 
      let processorSn := exist (fun n=>n<bsp_p) (S n) H in
        let processorN := exist (fun n=>n<bsp_p) n (Lt.lt_trans n (S n) bsp_p (lt_n_Sn n) H) in
          processorsUpTo processorSn = (processorsUpTo processorN) ++ (processorSn::nil).
  Proof.
    intros n H processorSn processorN; unfold processorsUpTo.
    assert(forall n0, In n0 (seq (S n) 1) -> n0 < bsp_p) as Hn by
      (intros n' H0; destruct H0; elim H0; assumption).
    assert(forall n0, In n0 ((seq 0 (S (proj1_sig processorN))) ++ (seq (S n) 1)) -> n0<bsp_p) as H3 by
      (intros n0 H0;simpl in H0; destruct H0; [
         rewrite <- H0; apply bsp_pNotZero |
         apply in_app_or in H0; destruct H0 as [H0 | H0]; [
           apply inSeqLt in H0; omega |
           simpl in H0; destruct H0; elim H0; assumption ] ]).
    assert(forall n':nat, In n' (seq 0 (S(S n))) -> n' < bsp_p) as H3' by
      (intros n' H0; apply H3; 
        rewrite <- seqAppN; rewrite plus_comm; assumption).
    replace (processorSn::nil) with (inSig _ (seq (S n) (S O)) Hn).
      rewrite inSig_app with (H3:=H3);
        rewrite inSig_eq with (H1:=H3)(H2:=H3')(l2:=seq 0 (S(S n))).
      apply inSig_eq; trivial.
      rewrite <- seqAppN; rewrite plus_comm; trivial.
      simpl; f_equal; apply proj1_sigBoundedNatInjective; trivial.
  Qed.

End TYPE.

Module Make  
  (BsmlPrimitives: PRIMITIVES) : 
  TYPE(BsmlPrimitives).

  Include TYPE(BsmlPrimitives).

End Make.