Require Export Bool.Bvector.
Require Import LIFO.Nat.BoundedNat.
Require Import Omega.

Section Functions_definitions.

  Variable A : Type.

  Fixpoint In (n:nat)(a:A)(v:vector A n) : Prop :=
  match v with 
    | Vnil => False
    | Vcons a' n' v' => a=a' \/ In n' a v'
  end.

  Fixpoint map (B:Type)(f:A->B)(n:nat)(v:vector A n) : 
    vector B n :=
    match v with
      | Vnil => Vnil B
      | Vcons a n' v' => Vcons _ (f a) n' (map B f n' v')
    end.

  Definition combine (B:Type) (n:nat) 
    (v1: vector A n) (v2: vector B n):
    vector (A*B) n.
  Proof.
    induction n.
      exact(Vnil (A*B)).    
      pose(a:=Vhead A n v1);
        pose(b:=Vhead B n v2);
          pose(va:=Vtail A n v1);
            pose(vb:=Vtail B n v2);
              exact (Vcons (A*B) (a,b) n (IHn va vb)).
  Defined.

  Definition nth (k:nat) (n :nat) (v:vector A n) (default: A)  : A.
  Proof.
    generalize dependent k.
    induction n.
      intros k ; exact default.
      intros k;
      pose(a:=Vhead A n v);
      pose(va:=Vtail A n v);
      induction k.
        exact a.
        exact (IHn va k).
  Defined.
  
  Fixpoint nth' (i:nat) (n:nat) (v:vector A n) : option A :=
  match i,v with 
    | O, (Vcons x _ _) => Some x
    | S i', (Vcons x n' rem) => nth' i' n' rem
    | _, _ => None
  end. 

  Lemma nth'Some:
    forall (n:nat)(v:vector A n)(i:nat),
    i < n -> exists a, nth' i n v = Some a.
  Proof.
    induction v; intros i Hi.
      contradict Hi; omega. 
      destruct i.
      simpl; exists a; auto.
      simpl; apply IHv; intuition.
  Qed.

  Definition strongNth (n:nat)(v:vector A n)(i:nat|i<n): 
    { a:A | nth' (proj1_sig i) n v = Some a }.
  Proof.
    induction v; destruct i as [i Hi].
      contradict Hi; omega.
      destruct i. 
        exists a; auto.
        pose(H:=Hi); apply lt_S_n in H.
        exact(IHv (exist _ i H)).
  Defined.

  Fixpoint seq (n:nat)(first:nat) : vector nat n :=
    match n as n' return vector nat n' with 
      | O   => Vnil nat 
      | S n => Vcons nat first n (seq n (S first))
    end.

  Lemma plusLt:
    forall(n m:nat), n < n+S m.
  Proof.
    intros; omega.
  Qed.

  Definition typeConv (n m : nat)(i:nat|i<S n +m):
    {i:nat | i<n+S m}.
  Proof.
    destruct i as [i Hi]; exists i.
    rewrite plus_comm; simpl; rewrite plus_comm; assumption.
  Defined.

  Definition to_list: forall (n:nat)(v:vector A n), 
    {l:List.list A | List.length l = n}.
  Proof.
    intros n v.
    induction n.
    exists List.nil. reflexivity.
    pose(a:=Vhead A n v).
    pose(va:=Vtail A n v).
    pose(la:= match IHn va with exist la Hla => la end).
    exists (List.cons a la).
    unfold la.
    case (IHn va).
    intros x H. simpl. rewrite H. reflexivity.
  Defined.

  Fixpoint for_all (P:A->Prop) (n:nat) (v:vector A n) : Prop := 
    match v with 
      | Vnil => True
      | Vcons a n' v' => (P a)/\for_all P n' v'
    end.

End Functions_definitions.

Implicit Arguments In [A].
Implicit Arguments map [A B n].
Implicit Arguments combine [A B].
Implicit Arguments nth [A].
Implicit Arguments nth' [A].
Implicit Arguments strongNth [A n].
Implicit Arguments to_list [A].
Implicit Arguments for_all [A].

Section Facts.
  
  Variable A : Type.

  Lemma eq_dec:
    forall (n:nat) 
    (eqelement_dec: forall (a b:A),{a=b}+{a<>b}) 
    (v1 v2: vector A n),
    {v1=v2}+{v1<>v2}.
  Proof.
    intros n eqelement_dec v1 v2.
    induction n.
    rewrite (V0_eq A v1).
    rewrite (V0_eq A v2).
    left. reflexivity.
    destruct (eqelement_dec (Vhead A n v1) (Vhead A n v2)) as [Heq | Hdiff].
    rewrite (VSn_eq A n v1).
    rewrite (VSn_eq A n v2).
    destruct (IHn (Vtail A n v1) (Vtail A n v2)) as [Hveq | Hvdiff].
      left. rewrite Hveq. rewrite Heq. reflexivity.
      right. 
        intro H.
        apply Hvdiff.
        rewrite (Vid_eq A (S n) v1).
        rewrite (Vid_eq A (S n) v2).
        unfold Vid. rewrite H. reflexivity.
      right. intro H. apply Hdiff. rewrite H. reflexivity.
  Qed.

  Lemma nth_Vnil (v:vector A 0) (default: A) (k:nat) : 
    nth k 0 v default = default.
  Proof.
    auto.
  Qed.

  Lemma nth_Vcons_O (n:nat) (v:vector A (S n)) (default: A)  : 
    nth O (S n) v default = Vhead A n v.
  Proof.
    auto.
  Qed.

  Lemma nth_Vcons_Sn (n:nat)(v:vector A (S n))(default: A)(k:nat): 
    nth (S k) (S n) v default = nth k n (Vtail A n v) default.
  Proof.
    auto.
  Qed.

  Lemma nth_nth' (n:nat) (v:vector A n):forall (k:nat),
    nth k n (map (fun (x:A)=>Some x) v) None = nth' k n v.
  Proof.
    induction v.
      induction k ; trivial.
      induction k. 
        trivial.
        apply IHv.
  Qed.

  Lemma nth_to_list (n:nat) (v:vector A n): 
    forall (k:nat) (default: A),
    nth k n v default = 
    List.nth k (projT1 (to_list n v)) default.
  Proof.  
    induction v.
      induction k; trivial.
      induction k ; intro default.
        reflexivity.
        simpl.
        rewrite IHv.
        f_equal.
        destruct (to_list n v).
        reflexivity.
  Qed.

  Lemma map_map:
    forall (B C : Type)
      (f : A -> B)(g : B -> C)(n:nat)(v : vector A n),
      map g (map f v) = map (fun x : A => g (f x)) v.
  Proof.
    induction v.
      auto.
      simpl; rewrite IHv; reflexivity.
  Qed.

  Lemma map_ext:
    forall (B : Type) (f g : A -> B),
      (forall a : A, f a = g a) -> 
      forall n (v : vector A n), map f v = map g v.
  Proof.
    induction v.
      auto.
      simpl; rewrite H; rewrite IHv ;reflexivity.
  Qed.

  Lemma proj1_sigStrongNthNth:
    forall(n:nat)(i:nat|i<n)(v:vector A n)(d:A),
      proj1_sig(strongNth v i) = nth (proj1_sig i) n v d.
  Proof.
    induction v; destruct i as [i Hi]. 
      contradict Hi; omega.
      destruct i.
        reflexivity.
        simpl; 
        apply (IHv(exist (fun i=>i<n) i (lt_S_n i n Hi))).
  Qed.

End Facts.

Lemma strongNthMap:
  forall(A B:Type)(f:A->B)(n:nat)(i:nat|i<n)(v:vector A n),
    proj1_sig(strongNth (map f v) i) = f(proj1_sig(strongNth v i)).
Proof.
  induction v; destruct i as [i Hi]. 
    contradict Hi; omega.
    destruct i.
      reflexivity.
      simpl; apply (IHv(exist (fun i=>i<n) i (lt_S_n i n Hi))).
Qed.

Lemma nthMap:
  forall(A B:Type)(f:A->B)(n:nat)(i:nat)(v:vector A n)(a:A),
    nth i n (map f v) (f a) = f(nth i n v a).
Proof.
  intros A B f n i v; generalize dependent i;  induction v.
    reflexivity. 
    destruct i.
      intro a0; auto.
      apply (IHv i).
Qed.

Fixpoint seqLt (n:nat)(first:nat) : vector {i:nat|i<first+n} n :=
  match n as n' return vector {i:nat|i<first+n'} n' with 
    | O   => Vnil {i:nat|i<first+0}      
    | S n => 
      let first':{i:nat|i<first+S n}:=
        exist _ first (plusLt first n) in
          Vcons _ first'  n 
          (map (typeConv first n) (seqLt n (S first)))
  end.

Program Definition 
  strongS (n first:nat)(i:nat|i<first+n):{i:nat|i<S first+ n} := 
  S i.
Next Obligation.
  apply lt_n_S; assumption.
Defined.

Lemma seqLtMapS:
  forall(n:nat)(first:nat),
    seqLt n (S first) = map (strongS n first) (seqLt n first).
Proof.
  induction n.
  auto.
  intro first; simpl; f_equal; unfold strongS; simpl.
    apply proj1_sigBoundedNatInjective; reflexivity.
    rewrite (IHn (S first));
      repeat rewrite map_map;
        apply map_ext;
          intros a;
            apply proj1_sigBoundedNatInjective;
              destruct a; reflexivity.
Qed.
  
Lemma vMapNthSeq: 
  forall (A:Type)(n:nat)(v:vector A n),
    v = map (fun i=>proj1_sig(strongNth v i)) (seqLt n 0).
Proof.
  induction v.
    auto.
    rewrite map_ext with (g:=fun (i:nat|i<S n)=>nth (proj1_sig i) (S n) (Vcons A a n v) a).
      simpl in *; f_equal; rewrite seqLtMapS; repeat rewrite map_map; rewrite IHv; apply map_ext;
      intro i; rewrite proj1_sigStrongNthNth with (d:=a).
          set(v':=(map (fun i0 : {i0 : nat | i0 < n} => proj1_sig (strongNth v i0))
            (seqLt n 0)));
          rewrite IHv. reflexivity.
      intro i. apply proj1_sigStrongNthNth with (d:=a).
Qed.      



Program Definition shift (n first:nat)(i:nat|i<n):{i:nat|i<first+n} :=
  first+i.
Next Obligation.
  omega.
Defined.

Lemma proj1_sigTypeConv: 
  forall (first n:nat)(i:nat|i<S first + n), 
    proj1_sig (typeConv first n i) = proj1_sig i.
Proof.
  intros first n i; destruct i; auto.
Qed.
  
Lemma strongNthSeqLtFirst:
  forall (n:nat)(first:nat)(i:nat|i<n),
    proj1_sig(strongNth(seqLt n first) i) = shift n first i.
Proof.
  induction n; intros first i; destruct i as [i Hi].
      contradict Hi; omega.
      destruct i.
        simpl; apply proj1_sigBoundedNatInjective; simpl; auto with arith.
        simpl; rewrite seqLtMapS;
          replace(map (strongS n first) (seqLt n first)) with (seqLt n (S first)).
            set(is:=((exist (fun i0 : nat => i0 < n) i (Lt.lt_S_n i n Hi))));
              rewrite strongNthMap with (i:=is);
                rewrite proj1_sigStrongNthNth with (d:=shift n (S first) is); 
                  pose(H:=IHn (S first) is);
                    rewrite proj1_sigStrongNthNth with (d:=shift n (S first) is) in H;
                      simpl in *; rewrite H;
                        apply proj1_sigBoundedNatInjective;
                          rewrite proj1_sigTypeConv;
                            simpl; intuition.
            induction n; auto;
              simpl; f_equal; [
                unfold strongS; apply proj1_sigBoundedNatInjective; auto |
                  rewrite seqLtMapS; 
                    repeat rewrite map_map;
                      apply map_ext;
                        intro x; apply proj1_sigBoundedNatInjective; 
                          unfold strongS; destruct x; auto].
Qed.

Lemma strongNthSeqLt0:
  forall (n:nat)(i:nat|i<n),
    proj1_sig(strongNth(seqLt n 0) i) = i.
Proof.
  intros n i;
    rewrite (strongNthSeqLtFirst n 0);
      unfold shift; apply proj1_sigBoundedNatInjective; auto.
Qed.

Lemma strongNthCombine: 
  forall (A B: Type)(n:nat)(v1: vector A n)(v2:vector B n)(i:nat|i<n), 
    proj1_sig(strongNth (combine n v1 v2) i) = 
    (proj1_sig (strongNth v1 i),proj1_sig (strongNth v2 i)).
Proof.
  intros A B n v1 v2 i;
    induction n; destruct i as [i Hi]. 
      contradict Hi; omega.
      rewrite VSn_eq with (v:=v1); rewrite VSn_eq with (v:=v2); destruct i.
        reflexivity.
        simpl; apply  IHn with (i:=(exist (fun i0 : nat => i0 < n) i (Lt.lt_S_n i n Hi))).
Qed.

Definition matrix(A:Type)(n m:nat) := vector (vector A m) n.

Definition 
  getRow(A:Type)(n m:nat)(mat:matrix A n m)(i:nat|i<m) :
  vector A n := 
  map(fun column =>proj1_sig(strongNth column i)) mat. 

Implicit Arguments getRow [A n m].

Definition transpose(A:Type)(n m:nat)(mat:matrix A n m) : 
  matrix A m n :=
  map(getRow mat)(seqLt m 0).

Implicit Arguments transpose [A n m].

Definition 
  get(A:Type)(n m:nat)(mat:matrix A n m)
  (col:nat|col<n)(row:nat|row<m): 
  A :=
  proj1_sig(strongNth(proj1_sig(strongNth mat col)) row).

Implicit Arguments get [A n m].

Ltac contradictBoundedNat := 
  match goal with 
    | [ n:{i:nat|i<0} |- _ ] =>
      let Hn' := fresh "H" in
        let n' := fresh "n" in
        destruct n as [n' Hn'];
          contradict Hn';
          omega
  end.

Lemma transposeProperty: 
  forall(A:Type)(n m:nat)(mat:matrix A n m)
    (col:nat|col<n)(row:nat|row<m),
    get mat col row = get (transpose mat) row col.
Proof.
  intros A n m mat col row;
    unfold transpose, get, getRow;
      repeat rewrite strongNthMap;
        replace (@proj1_sig (@sig nat (fun i : nat => lt i m))
          (fun a : @sig nat (fun i : nat => lt i m) =>
            @eq (option (@sig nat (fun i : nat => lt i m)))
            (@nth' (@sig nat (fun i : nat => lt i m))
              (@proj1_sig nat (fun i : nat => lt i m) row) m
              (seqLt m O)) (@Some (@sig nat (fun i : nat => lt i m)) a))
          (@strongNth (@sig nat (fun i : nat => lt i m)) m 
            (seqLt m O) row)) with row 
  by (symmetry; apply strongNthSeqLt0);
  reflexivity.
Qed.
